import base64, re
from PIL import Image
from io import BytesIO

def saveToBmp(fileIn,fileOut,width=None,fmt="BMP",base64Input=False):
    if (type(fileIn) == str and fileIn.startswith("data:image")) or base64Input:
        b64 = re.sub('^data:image/.+;base64,', '', fileIn)
        image_data = base64.b64decode(b64)
        img = Image.open(BytesIO(image_data))
    else:    
        img = Image.open(fileIn)


    w, h = img.size
    if h > 128 or w > 160 or fmt.upper() == 'BMP':
        img = img.convert("RGB")

    if h > 128 or w > 160:
        # do resize
        ratio_h = 128/h
        ratio_w = 160/w
        
        _w = int(round(w*ratio_h, 0))
        _h = int(round(h*ratio_h, 0))
        if _w<160:
            _w = int(round(w*ratio_w, 0))
            _h = int(round(h*ratio_w, 0))
        # print("resize img {} {}".format((w,h), (_w,_h)))
        img = img.resize((_w,_h))

    img.save(fileOut, fmt.upper())

if __name__ == "__main__":
    from hashlib import md5
    saveToBmp('down.png', 'down_cvt.png', fmt='png')
    print("md5 oss:", md5(open('down_oss.png', "rb").read()).hexdigest())
    print("md5 converted:", md5(open('down_cvt.png', "rb").read()).hexdigest())
    # encoded = base64.b64encode(data)
    # saveToBmp("data:image/jpeg;base64,"+encoded.decode(), 'test2.bmp')




# -*- coding: utf-8 -*-
"""
This module contains functions for turning a Python script into a .hex file
and flashing it onto a BBC micro:bit.

Copyright (c) 2015-2018 Nicholas H.Tollervey and others.

See the LICENSE file for more information, or visit:

https://opensource.org/licenses/MIT
"""

from __future__ import print_function

import argparse
import binascii
import ctypes
import os
import struct
import sys
import shutil
import base64, re
from io import BytesIO

from subprocess import check_output
import time

# nudatus is an optional dependancy
can_minify = True
try:
    import nudatus
except ImportError:  # pragma: no cover
    can_minify = False

#: The magic start address in flash memory for a Python script.
_SCRIPT_ADDR = 0x3e000


#: The help text to be shown when requested.
_HELP_TEXT = """
Flash Python onto the BBC micro:bit or extract Python from a .hex file.

If no path to the micro:bit is provided uflash will attempt to autodetect the
correct path to the device. If no path to the Python script is provided uflash
will flash the unmodified MicroPython firmware onto the device. Use the -e flag
to recover a Python script from a hex file. Use the -r flag to specify a custom
version of the MicroPython runtime.

Documentation is here: https://uflash.readthedocs.io/en/latest/
"""


#: MAJOR, MINOR, RELEASE, STATUS [alpha, beta, final], VERSION of uflash
_VERSION = (1, 2, 4, )
_MAX_SIZE = 8188


#: The version number reported by the bundled MicroPython in os.uname().
MICROPYTHON_VERSION = '1.0.1'


def get_version():
    """
    Returns a string representation of the version information of this project.
    """
    return '.'.join([str(i) for i in _VERSION])


def get_minifier():
    """
    Report the minifier will be used when minify=True
    """
    if can_minify:
        return 'nudatus'
    return None


def strfunc(raw):
    """
    Compatibility for 2 & 3 str()
    """
    return str(raw) if sys.version_info[0] == 2 else str(raw, 'utf-8')


def hexlify(script, minify=False):
    """
    Takes the byte content of a Python script and returns a hex encoded
    version of it.

    Based on the hexlify script in the microbit-micropython repository.
    """
    if not script:
        return ''
    # Convert line endings in case the file was created on Windows.
    script = script.replace(b'\r\n', b'\n')
    script = script.replace(b'\r', b'\n')
    if minify:
        if not can_minify:
            raise ValueError("No minifier is available")
        script = nudatus.mangle(script.decode('utf-8')).encode('utf-8')
    # Add header, pad to multiple of 16 bytes.
    data = b'MP' + struct.pack('<H', len(script)) + script
    # Padding with null bytes in a 2/3 compatible way
    data = data + (b'\x00' * (16 - len(data) % 16))
    if len(data) > _MAX_SIZE:
        # 'MP' = 2 bytes, script length is another 2 bytes.
        raise ValueError("Python script must be less than 8188 bytes.")
    # Convert to .hex format.
    output = [':020000040003F7']  # extended linear address, 0x0003.
    addr = _SCRIPT_ADDR
    for i in range(0, len(data), 16):
        chunk = data[i:min(i + 16, len(data))]
        chunk = struct.pack('>BHB', len(chunk), addr & 0xffff, 0) + chunk
        checksum = (-(sum(bytearray(chunk)))) & 0xff
        hexline = ':%s%02X' % (strfunc(binascii.hexlify(chunk)).upper(),
                               checksum)
        output.append(hexline)
        addr += 16
    return '\n'.join(output)


def unhexlify(blob):
    """
    Takes a hexlified script and turns it back into a string of Python code.
    """
    lines = blob.split('\n')[1:]
    output = []
    for line in lines:
        # Discard the address, length etc. and reverse the hexlification
        output.append(binascii.unhexlify(line[9:-2]))
    # Check the header is correct ("MP<size>")
    if (output[0][0:2].decode('utf-8') != u'MP'):
        return ''
    # Strip off header
    output[0] = output[0][4:]
    # and strip any null bytes from the end
    output[-1] = output[-1].strip(b'\x00')
    script = b''.join(output)
    try:
        result = script.decode('utf-8')
        return result
    except UnicodeDecodeError:
        # Return an empty string because in certain rare circumstances (where
        # the source hex doesn't include any embedded Python code) this
        # function may be passed in "raw" bytes from MicroPython.
        return ''


def embed_hex(runtime_hex, python_hex=None):
    """
    Given a string representing the MicroPython runtime hex, will embed a
    string representing a hex encoded Python script into it.

    Returns a string representation of the resulting combination.

    Will raise a ValueError if the runtime_hex is missing.

    If the python_hex is missing, it will return the unmodified runtime_hex.
    """
    if not runtime_hex:
        raise ValueError('MicroPython runtime hex required.')
    if not python_hex:
        return runtime_hex
    py_list = python_hex.split()
    runtime_list = runtime_hex.split()
    embedded_list = []
    # The embedded list should be the original runtime with the Python based
    # hex embedded two lines from the end.
    embedded_list.extend(runtime_list[:-5])
    embedded_list.extend(py_list)
    embedded_list.extend(runtime_list[-5:])
    return '\n'.join(embedded_list) + '\n'


def extract_script(embedded_hex):
    """
    Given a hex file containing the MicroPython runtime and an embedded Python
    script, will extract the original Python script.

    Returns a string containing the original embedded script.
    """
    hex_lines = embedded_hex.split('\n')
    script_addr_high = hex((_SCRIPT_ADDR >> 16) & 0xffff)[2:].upper().zfill(4)
    script_addr_low = hex(_SCRIPT_ADDR & 0xffff)[2:].upper().zfill(4)
    start_script = None
    within_range = False
    # Look for the script start address
    for loc, val in enumerate(hex_lines):
        if val[0:9] == ':02000004':
            # Reached an extended address record, check if within script range
            within_range = val[9:13].upper() == script_addr_high
        elif within_range and val[0:3] == ':10' and \
                val[3:7].upper() == script_addr_low:
            start_script = loc
            break
    if start_script:
        # Find the end of the script
        end_script = None
        for loc, val in enumerate(hex_lines[start_script:]):
            if val[9:41] == 'FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF':
                end_script = loc + start_script
                break
        # Pass the extracted hex through unhexlify
        return unhexlify('\n'.join(
            hex_lines[start_script - 1:end_script if end_script else -6]))
    return ''


def find_device(preferdisk=None, sizeCheck=None):
    """
    Returns a path on the filesystem that represents the plugged in BBC
    micro:bit that is to be flashed. If no micro:bit is found, it returns
    None.

    Works on Linux, OSX and Windows. Will raise a NotImplementedError
    exception if run on any other operating system.
    """
    # Check what sort of operating system we're on.
    if os.name == 'posix':
        # 'posix' means we're on Linux or OSX (Mac).
        # Call the unix "mount" command to list the mounted volumes.
        mount_output = check_output('mount').splitlines()
        mounted_volumes = [x.split()[2] for x in mount_output]
        
        for volume in mounted_volumes:
            if preferdisk:
                if type(preferdisk) is str:
                    preferdisk = preferdisk.encode('utf8')
                if volume.endswith(preferdisk):
                    return volume.decode('utf-8')
            else:
                if volume.endswith(b'MICROBIT'):
                    return volume.decode('utf-8')  # Return a string not bytes.
                elif volume.endswith(b'ARCADE-F4'):
                    return volume.decode('utf-8')  # Return a string not bytes.
                elif volume.endswith(b'NanoBit'):
                    return volume.decode('utf-8')  # Return a string not bytes.
    elif os.name == 'nt':
        # 'nt' means we're on Windows.

        def get_volume_name(disk_name):
            """
            Each disk or external device connected to windows has an attribute
            called "volume name". This function returns the volume name for
            the given disk/device.

            Code from http://stackoverflow.com/a/12056414
            """
            vol_name_buf = ctypes.create_unicode_buffer(1024)
            serial_number = None
            max_component_length = None
            file_system_flags = None
            ctypes.windll.kernel32.GetVolumeInformationW(
                ctypes.c_wchar_p(disk_name), vol_name_buf,
                ctypes.sizeof(vol_name_buf), serial_number, max_component_length, None, None, 0)
            return vol_name_buf.value

        #
        # In certain circumstances, volumes are allocated to USB
        # storage devices which cause a Windows popup to raise if their
        # volume contains no media. Wrapping the check in SetErrorMode
        # with SEM_FAILCRITICALERRORS (1) prevents this popup.
        #
        old_mode = ctypes.windll.kernel32.SetErrorMode(1)
        defaultPathList = ['MICROBIT', 'ARCADE-F4', 'NanoBit']
        try:
            for disk in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
                path = '{}:\\'.format(disk)
                #
                # Don't bother looking if the drive isn't removable
                #
                if ctypes.windll.kernel32.GetDriveTypeW(path) != 2:
                    continue
                if os.path.exists(path):
                    pathName = get_volume_name(path)
                    if pathName in defaultPathList:
                        return path
                    elif sizeCheck != None:
                        if type(sizeCheck) == str:
                            sizeCheck = eval(sizeCheck.replace('M', '*1024*1024'))
                        total, used, free = shutil.disk_usage(path)
                        if abs(total - sizeCheck) < 0x20000:
                            return path
                    elif preferdisk != None:
                        if preferdisk == pathName:
                            return path
        finally:
            ctypes.windll.kernel32.SetErrorMode(old_mode)
    else:
        # No support for unknown operating systems.
        raise NotImplementedError('OS "{}" not supported.'.format(os.name))


def save_hex(hex_file, path):
    """
    Given a string representation of a hex file, this function copies it to
    the specified path thus causing the device mounted at that point to be
    flashed.

    If the hex_file is empty it will raise a ValueError.

    If the filename at the end of the path does not end in '.hex' it will raise
    a ValueError.
    """
    if not hex_file:
        return (-1, 'Cannot flash an empty .hex file.')
    if not path.endswith('.hex'):
        return (-1, 'The path to flash must be for a .hex file.')
    with open(path, 'wb') as output:
        output.write(hex_file.encode('ascii'))
        return (0, path)


def flash(path_to_python=None, paths_to_microbits=None,
          path_to_runtime=None, python_script=None, minify=False, echo=None):
    """
    Given a path to or source of a Python file will attempt to create a hex
    file and then flash it onto the referenced BBC micro:bit.

    If the path_to_python & python_script are unspecified it will simply flash
    the unmodified MicroPython runtime onto the device.

    If used, the python_script argument should be a bytes object representing
    a UTF-8 encoded string. For example::

        script = "from microbit import *\\ndisplay.scroll('Hello, World!')"
        uflash.flash(python_script=script.encode('utf-8'))

    If paths_to_microbits is unspecified it will attempt to find the device's
    path on the filesystem automatically.

    If the path_to_runtime is unspecified it will use the built in version of
    the MicroPython runtime. This feature is useful if a custom build of
    MicroPython is available.

    If the automatic discovery fails, then it will raise an IOError.
    """
    # Check for the correct version of Python.
    if not ((sys.version_info[0] == 3 and sys.version_info[1] >= 3) or
            (sys.version_info[0] == 2 and sys.version_info[1] >= 7)):
        raise RuntimeError('Will only run on Python 2.7, or 3.3 and later.')
    # Grab the Python script (if needed).
    python_hex = ''
    if path_to_python:
        if not path_to_python.endswith('.py'):
            return (-4, 'Python files must end in ".py".')
        with open(path_to_python, 'rb') as python_script:
            python_hex = hexlify(python_script.read(), minify)
    elif python_script:
        python_hex = hexlify(python_script, minify)

    # runtime = _RUNTIME
    # Load the hex for the runtime.
    if path_to_runtime:
        with open(path_to_runtime) as runtime_file:
            runtime = runtime_file.read()
    # Generate the resulting hex file.
    micropython_hex = embed_hex(runtime, python_hex)
    # Find the micro:bit.
    if not paths_to_microbits:
        found_microbit = find_device()
        if found_microbit:
            paths_to_microbits = [found_microbit]
    # Attempt to write the hex file to the micro:bit.
    if paths_to_microbits:
        for path in paths_to_microbits:
            hex_path = os.path.join(path, 'micropython.hex')
            print('Flashing Python to: {}'.format(hex_path))
            if echo:
                echo('Flashing Python to: {}'.format(hex_path))
            return save_hex(micropython_hex, hex_path)
    else:
        return (-2, 'Unable to find micro:bit. Is it plugged in?')

def copyfileobj(fsrc, fdst, callback, total, length=4*1024):
    copied = 0
    while True:
        buf = fsrc.read(length)
        if not buf:
            break
        fdst.write(buf)
        copied += len(buf)
        callback(copied, total)
        # this is for mac, cache ENXIO in runtime
        fdst.flush()
        os.fsync(fdst)

def copyHex(path_to_hex, callback):
    # todo: support uf2 borads other than microbits
    
    paths_to_microbits = None
    
    found_microbit = find_device()
    if found_microbit:
        paths_to_microbits = [found_microbit]
        if paths_to_microbits:
            for path in paths_to_microbits:
                hex_path = os.path.join(path, 'micropython.hex')
                # print('Flashing Firmware to: {}'.format(path))
                # shutil.copyfile(path_to_hex, hex_path)
                if type(path_to_hex) == bytes:
                    fsrc = BytesIO(path_to_hex)
                    size = len(path_to_hex)
                elif path_to_hex.startswith("data:"):
                    if path_to_hex.startswith("data:application"):
                        b64 = re.sub('^data:application/.+;base64,', '', path_to_hex)
                    else:
                        b64 = re.sub('^data:text/.+;base64,', '', path_to_hex)
                    b64Data = base64.b64decode(b64)
                    fsrc = BytesIO(b64Data)
                    size = len(b64Data)
                else:
                    fsrc = open(path_to_hex, 'rb')
                    size = os.stat(path_to_hex).st_size
                with open(hex_path, 'wb') as fdst:
                    copyfileobj(fsrc, fdst, callback, size)
            return (0, paths_to_microbits[0])
        else:
            return (-1, 'Unable to find target')
    else:
        return (-1, 'Unable to find target')

def getDisk(disk_filter, size=None):
    found_microbit = find_device(disk_filter, size)
    return found_microbit

def extract(path_to_hex, output_path=None):
    """
    Given a path_to_hex file this function will attempt to extract the
    embedded script from it and save it either to output_path or stdout
    """
    with open(path_to_hex, 'r') as hex_file:
        python_script = extract_script(hex_file.read())
        if output_path:
            with open(output_path, 'w') as output_file:
                output_file.write(python_script)
        else:
            print(python_script)


def watch_file(path, func, *args, **kwargs):
    """
    Watch a file for changes by polling its last modification time. Call the
    provided function with *args and **kwargs upon modification.
    """
    if not path:
        raise ValueError('Please specify a file to watch')
    print('Watching "{}" for changes'.format(path))
    last_modification_time = os.path.getmtime(path)
    try:
        while True:
            time.sleep(1)
            new_modification_time = os.path.getmtime(path)
            if new_modification_time == last_modification_time:
                continue
            func(*args, **kwargs)
            last_modification_time = new_modification_time
    except KeyboardInterrupt:
        pass


# def main(argv=None):
#     """
#     Entry point for the command line tool 'uflash'.

#     Will print help text if the optional first argument is "help". Otherwise
#     it will ensure the optional first argument ends in ".py" (the source
#     Python script).

#     An optional second argument is used to reference the path to the micro:bit
#     device. Any more arguments are ignored.

#     Exceptions are caught and printed for the user.
#     """
#     if not argv:
#         argv = sys.argv[1:]

#     parser = argparse.ArgumentParser(description=_HELP_TEXT)
#     parser.add_argument('source', nargs='?', default=None)
#     parser.add_argument('target', nargs='*', default=None)
#     parser.add_argument('-r', '--runtime', default=None,
#                         help="Use the referenced MicroPython runtime.")
#     parser.add_argument('-e', '--extract',
#                         action='store_true',
#                         help=("Extract python source from a hex file"
#                               " instead of creating the hex file."), )
#     parser.add_argument('-w', '--watch',
#                         action='store_true',
#                         help='Watch the source file for changes.')
#     parser.add_argument('-m', '--minify',
#                         action='store_true',
#                         help='Minify the source')
#     parser.add_argument('--version', action='version',
#                         version='%(prog)s ' + get_version())
#     parser.add_argument('-u', '--uf2',
#                         action='store_true',
#                         help='UF2 like hex copy.')
#     parser.add_argument('-d', '--disk',
#                         action='store_true',
#                         help='Get Disk.')                    
#     args = parser.parse_args(argv)

#     if args.extract:
#         try:
#             extract(args.source, args.target)
#         except Exception as ex:
#             error_message = "Error extracting {source}: {error!s}"
#             print(error_message.format(source=args.source, error=ex),
#                   file=sys.stderr)
#             sys.exit(1)

#     elif args.watch:
#         try:
#             watch_file(args.source, flash,
#                        path_to_python=args.source,
#                        paths_to_microbits=args.target,
#                        path_to_runtime=args.runtime)
#         except Exception as ex:
#             error_message = "Error watching {source}: {error!s}"
#             print(error_message.format(source=args.source, error=ex),
#                   file=sys.stderr)
#             sys.exit(1)
#     elif args.uf2:
#         copyHex(args.source)
#     elif args.disk:
#         getDisk(args.source)
#     else:
#         try:
#             flash(path_to_python=args.source, paths_to_microbits=args.target,
#                   path_to_runtime=args.runtime, minify=args.minify)
#         except Exception as ex:
#             error_message = (
#                 "Error flashing {source} to {target}{runtime}: {error!s}"
#             )
#             source = args.source
#             target = args.target if args.target else "microbit"
#             if args.runtime:
#                 runtime = "with runtime {runtime}".format(runtime=args.runtime)
#             else:
#                 runtime = ""
#             print(error_message.format(source=source, target=target,
#                                        runtime=runtime, error=ex),
#                   file=sys.stderr)
#             sys.exit(1)



import sys, os, platform
from .ExecThread import *
from .esptool import main as espmain

def esptoolWork(commmand, onLog, onFinish, cdc=None):
    cmd = commmand.split(' ')
    def kprint(*args, **kwargs):
        onLog(args[0])
    try:
        ret = espmain(cmd, kprint, cdc)
        onFinish(0)
    except Exception as e:
        onFinish("ESPTOOL: "+str(e))

def uploadEsp32Firm(port, config, onLog, onFinish, cdc=None):
    if port:
        port = '--port '+port
    
    # esptool.main(commmand)
    cwd = os.getcwd()
    firmPath = os.path.join(cwd, config['firmwarePath'])
    flashAddr = config['flashAddr']

    commmand = '--chip esp32 {} --baud 1024000 write_flash -z --flash_mode dio --flash_freq 40m {} {}'.format(port, flashAddr, firmPath)
    th = threading.Thread(target=esptoolWork, args=(commmand, onLog, onFinish, cdc, ))
    th.daemon = True
    th.start()
    




# -*- coding:utf-8 -*-

import subprocess, signal, platform
import os
import threading
import locale

class ExecThread(threading.Thread):
    def __init__(self, cmd, cwd, outputCb, onFinish, extraEnv={}):
        self.running = True
        self.cmdList = []
        if type(cmd) == list:
            self.cmdList = cmd
        else:
            self.cmdList = [cmd]
        self.cwd = cwd
        self.env = {**os.environ, **extraEnv}
        self.onFinish = onFinish
        self.outputCb = outputCb
        threading.Thread.__init__(self)

    def run(self):
        try:
            while len(self.cmdList) > 0 and self.running:
                cmd = self.cmdList.pop(0)
                if platform.system() == "Windows":
                    self.process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=self.cwd, shell=True, env=self.env)
                else:
                    self.process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=self.cwd,
                                                    shell=True, preexec_fn=os.setsid, env=self.env)

                while self.process.poll() is None:
                    line = self.process.stdout.readline()
                    if self.outputCb:
                        self.outputCb(line)
                    # else:
                    #     print('>>', ret)
                # print('exec done',self.process.poll())
            if self.onFinish:
                self.onFinish(self.process.poll())
        except Exception as e:
            if self.onFinish:
                self.onFinish(e)

    def killTh(self):
        if self.process.returncode == None:
            # print("kill ", self.process.pid, self.process.returncode)
            self.running = False
            if platform.system() == "Windows":
                p = subprocess.Popen("taskkill /F /T /PID %i" % self.process.pid, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
                while p.poll() is None:
                    line = p.stdout.readline()
                    # print('>>', line)
            else:
                os.killpg(os.getpgid(self.process.pid), signal.SIGKILL)

if __name__ == "__main__":
    # for test
    # cmd = "pio run -t upload"
    cmd = [
        "pio init --board uno", 
        "pio run -t upload"
    ]
    env = {
        "PATH": "E:/kittenblock/python/win/Scripts",
        "PLATFORMIO_CORE_DIR": "E:/kittenblock/python/pio"
    }
    cwd = "E:/kittenblock/python/test"
    # def onOutput(msg):
    #     print(">>", msg)
    # def onFinish(code):
    #     print("Done %s" %code)
    th = ExecThread(cmd, cwd, onOutput, onFinish, env)
    th.setDaemon(True)
    th.start()

    a = input()



# -*- coding:utf-8 -*-
import sys, os
import json
import time
import threading
import struct
import base64
import shutil
import platform
from flask import request

from .ExecThread import *
from .MessageHandler import handle
from .Uf2Manager import *
from .SerialCom import serialList, serialCom
from .uflash import getDisk
from .ImageManager import saveToBmp

extensions = {
    'meowbit': {
        "id": "meowbit",
        "fullpath": os.getcwd(),
        "name": "Meowbit",
        "name_zh": "喵比特",
        "version": "1.0.0",
        "type": "micropy",
        "fwtype": "uf2",
        "replmode": True,
        "execmode": True,
        "firmware": "firmware.uf2",
        "thumbdisk": "PYBFLASH",
        "pyb": False
    }, 
    'meow32': {
        "id": "meow32",
        "fullpath": os.getcwd(),
        "name": "Meow32",
        "name_zh": "喵先生",
        "version": "1.0.0",
        "type": "micropy",
        "fwtype": "esptool",
        "replmode": True,
        "execmode": True,
        "pyb": True
    }
}
userPath = None
pioPath = None
server = None
uihandler = None
cloudVar = {}
cloudClients = {}

# def enumExtensionDir(args):
#     global userPath
#     userPath = args.user
#     if not userPath:
#         userPath = os.getcwd()
#     print("User Path: %s" %userPath)

def echo_socket(ws):
    handle(ws, extensions, userPath)

# @app.route('/test/<case>/<path:filename>')
# def loadTestCase(case, filename):
#     return send_from_directory(os.path.join(os.getcwd(), 'test', case), filename)

def installdriver():
    if platform.system() == "Windows" and platform.release() == '7':
        # infPath = os.path.join(userPath, "pybcdc.inf")
        # if platform.machine().endswith('64'):
        #     n = os.system('%%SystemRoot%%\\Sysnative\\pnputil.exe -i -a %s' %(infPath))
        # else:
        #     n = os.system('pnputil.exe -i -a %s' %(infPath))
        return str(0)
    else:
        return "只有windows7用户才需要装驱动哦～"

# @app.route('/readdisk/<prompt>/<path:filename>')
# def readdisk(prompt, filename):
#     try:
#         n = getDisk(prompt)
#         a = open(os.path.join(n, filename),"r").read()
#         return a
#     except Exception as err:
#         return '',403

def waitdisk(prompt, retry=1):
    sz = request.args.get('size')
    timeoutCnt = int(retry)
    while timeoutCnt > 0:
        timeoutCnt-=1
        n = getDisk(prompt, sz)
        if n:
            return n
        if timeoutCnt > 0:
            time.sleep(1)
    return "Cannot find disk", 403

def run(app, sockets, prefix='/hardware'):
    print('klink run')
    global userPath
    if not userPath:
        userPath = os.getcwd()
    sockets.add_url_rule(prefix, 'echo_socket', echo_socket)
    app.add_url_rule(prefix + '/installdriver', 'installdriver', installdriver)
    app.add_url_rule(prefix + '/waitdisk/<prompt>/<retry>', 'waitdisk', waitdisk)
    



import subprocess, signal, platform
import sys, os, time
import json
import logging
import threading
import locale
import shutil
import base64, re, requests
from io import BytesIO
from .kerror import KERR
from .ExecThread import *
from .uflash import copyHex, flash, getDisk

def uploadFirmTask(firmPath, ext, sendReq, onDone):
    def uploadProgress(p, total, prompt="#"):
        progress = int(p/total*100)
        sendReq("upload-status", {"text": prompt, "progress": progress})
    if firmPath.startswith("http"):
        firmPath = requests.get(firmPath).content
    try:
        

        (code, ret) = copyHex(firmPath, callback=uploadProgress)
        
        print("upload return %s %s" %(code, ret))
        newVol = None
        sendReq("upload-status")
        if code != 0:
            print('code != 0 error')
            onDone(ret, 1)
        else:
            # wait new thumbdisk show up
            if 'thumbdisk' in ext:
                DISKNAME = ext['thumbdisk']
                while not newVol:
                    newVol = getDisk(DISKNAME)
                    time.sleep(0.5)
            onDone(newVol,0)
    except Exception as err:
        print('exception error')
        onDone(str(err), 1)

def uploadUf2(ext, params, sendReq, userPath, onDone):

    sendReq("upload-status", {"text": "Initializing Upload", "progress": "1"})
    if "code" in params:
        if "boardName" in params:
            firmware = ext["board"][params["boardName"]]["runtime"]
            firmPath = os.path.join(ext['fullpath'], firmware)
        else:
            firmPath = os.path.join(ext['fullpath'], ext['runtime'])
        logging.info("upload %s" %firmPath)
        pycode = bytes(params['code'], encoding = "utf8")
        (code, ret) = flash(path_to_runtime=firmPath, python_script=pycode)
        sendReq("upload-status")
    else:
        if "base64" in params:
            firmPath = params['base64']
        elif "url" in params:
            firmPath = params['url']
        else:
            firmPath = os.path.join(ext['fullpath'], ext['firmware'])
        th = threading.Thread(target=uploadFirmTask,args=(firmPath, ext, sendReq, onDone))
        th.start()
        
    


from .MeowbitHandler import MeowbitHandler

class HardwareHandlerFactory:
  def handle(self, hwVersion, websocket, extensions, userPath, tag=""):
    if hwVersion == 'meowbit':
      return MeowbitHandler(websocket, extensions, userPath, tag)



#!/usr/bin/env python

"""
pyboard interface
This module provides the Pyboard class, used to communicate with and
control the pyboard over a serial USB connection.
Example usage:
    import pyboard
    pyb = pyboard.Pyboard('/dev/ttyACM0')
Or:
    pyb = pyboard.Pyboard('192.168.1.1')
Then:
    pyb.enter_raw_repl()
    pyb.exec('pyb.LED(1).on()')
    pyb.exit_raw_repl()
Note: if using Python2 then pyb.exec must be written as pyb.exec_.
To run a script from the local machine on the board and print out the results:
    import pyboard
    pyboard.execfile('test.py', device='/dev/ttyACM0')
This script can also be run directly.  To execute a local script, use:
    ./pyboard.py test.py
Or:
    python pyboard.py test.py
"""

import sys
import time

_rawdelay = None
DEBUG = True

try:
    stdout = sys.stdout.buffer
except AttributeError:
    # Python2 doesn't have buffer attr
    stdout = sys.stdout

def stdout_write_bytes(b):
    b = b.replace(b"\x04", b"")
    stdout.write(b)
    stdout.flush()

def ticks(t0=0):
    if t0:
        return round(time.time()*1000) - t0
    return round(time.time()*1000)

class PyboardError(Exception):
    pass

class TelnetToSerial:
    def __init__(self, ip, user, password, read_timeout=None):
        import telnetlib
        self.tn = telnetlib.Telnet(ip, timeout=15)
        self.read_timeout = read_timeout
        if b'Login as:' in self.tn.read_until(b'Login as:', timeout=read_timeout):
            self.tn.write(bytes(user, 'ascii') + b"\r\n")

            if b'Password:' in self.tn.read_until(b'Password:', timeout=read_timeout):
                # needed because of internal implementation details of the telnet server
                time.sleep(0.2)
                self.tn.write(bytes(password, 'ascii') + b"\r\n")

                if b'for more information.' in self.tn.read_until(b'Type "help()" for more information.', timeout=read_timeout):
                    # login succesful
                    from collections import deque
                    self.fifo = deque()
                    return

        raise PyboardError('Failed to establish a telnet connection with the board')

    def __del__(self):
        self.close()

    def close(self):
        try:
            self.tn.close()
        except:
            # the telnet object might not exist yet, so ignore this one
            pass

    def read(self, size=1):
        while len(self.fifo) < size:
            timeout_count = 0
            data = self.tn.read_eager()
            if len(data):
                self.fifo.extend(data)
                timeout_count = 0
            else:
                time.sleep(0.25)
                if self.read_timeout is not None and timeout_count > 4 * self.read_timeout:
                    break
                timeout_count += 1

        data = b''
        while len(data) < size and len(self.fifo) > 0:
            data += bytes([self.fifo.popleft()])
        return data

    def write(self, data):
        self.tn.write(data)
        return len(data)

    def inWaiting(self):
        n_waiting = len(self.fifo)
        if not n_waiting:
            data = self.tn.read_eager()
            self.fifo.extend(data)
            return len(data)
        else:
            return n_waiting

class Pyboard:
    def __init__(self, serial, baudrate=115200, user='micro', password='python', wait=0, rawdelay=0):
        global _rawdelay
        _rawdelay = rawdelay
        self.serial = serial

    def close(self):
        ""
        # setattr(self.serial, 'pybMutex', False)

    def read_until(self, min_num_bytes, ending, timeout=500, data_consumer=None):
        t0 = ticks()
        if DEBUG: print("\033[0;36m read_until: {} \033[0m".format(ending))
        data = self.serial.read(min_num_bytes)
        if data_consumer:
            data_consumer(data)
        buffered_count = 0
        while True:
            if data.endswith(ending):
                break
            elif self.serial.inWaiting() > 0:
                if buffered_count > 10240: # esp32 如果卡在bootloader保证退出
                    return data
                new_data = self.serial.read(1)
                data = data + new_data
                if data_consumer:
                    data_consumer(new_data)
                buffered_count += 1
            else:
                if not self.serial.opened:
                    raise PyboardError("Port closed")
                if timeout is not None and ticks(t0) > timeout:
                    if DEBUG: print("\033[0;31m read until timeout {} \033[0m".format(ticks(t0)))
                    break
                time.sleep(0.01)
        if DEBUG: print("\033[0;36m data: {} \033[0m".format(data))
        return data

    def enter_normal_repl(self):
        # Brief delay before sending RAW MODE char if requests
        if _rawdelay > 0:
            time.sleep(_rawdelay)
        time.sleep(3.5)
        # ctrl-C twice: interrupt any running program
        self.serial.write(b'\r\x02\x03')
        time.sleep(0.1)
        self.serial.write(b'\x03')
        time.sleep(0.1)
        # flush input (without relying on serial.flushInput())
        n = self.serial.inWaiting()
        timeout = 0
        while n > 0:
            self.serial.read(n)
            n = self.serial.inWaiting()
            timeout+=1
            if timeout > 100:
                raise PyboardError('could not enter normal repl(time out)', n)
        self.serial.write(b'\x03')
        data = self.read_until(1, b'>>> ')
        if not data.endswith(b'>>> '):
            data = data if len(data)<10 else data[-9:]
            raise PyboardError('could not enter normal repl(no >>>)',data)
        n = self.serial.inWaiting()
        timeout = 0
        while n > 0:
            n = self.serial.inWaiting()
            self.serial.read(n)

    def enter_raw_repl(self):
        # Brief delay before sending RAW MODE char if requests
        if _rawdelay > 0:
            time.sleep(_rawdelay)

        rawReplRetry = 5
        while rawReplRetry:
            # ctrl-C twice: interrupt any running program
            self.serial.write(b'\r\x03')
            time.sleep(0.1)
            self.serial.write(b'\x03')
            time.sleep(0.1)

            # flush input (without relying on serial.flushInput())
            n = self.serial.inWaiting()
            timeout = 0
            while n > 0:
                self.serial.read(n)
                n = self.serial.inWaiting()
                timeout+=1
                if timeout > 100:
                    raise PyboardError('could not enter raw repl(time out)', n)
            # riven, no soft reboot on repl interaction
        
            self.serial.write(b'\r\x01') # ctrl-A: enter raw REPL
            data = self.read_until(1, b'raw REPL; CTRL-B to exit\r\n', timeout=100)
            if not data.endswith(b'raw REPL; CTRL-B to exit\r\n'):
                if DEBUG: print("\033[0;31m retry {} \033[0m".format(rawReplRetry))
                rawReplRetry -= 1
                if rawReplRetry == 0:
                    data = data if len(data)<10 else data[-9:]
                    if DEBUG: print("\033[0;36m could not enter raw repl \033[0m")
                    raise PyboardError('could not enter raw repl(no raw REPL; CTRL-B to exit)',data)
            else:
                break

    def exit_raw_repl(self):
        self.serial.write(b'\r\x02') # ctrl-B: enter friendly REPL
        time.sleep(0.1)

    def follow(self, timeout, data_consumer=None):
        # wait for normal output
        data = self.read_until(1, b'\x04', timeout=timeout, data_consumer=data_consumer)
        # print("<<<<", data)
        if not data.endswith(b'\x04'):
            data = data if len(data)<10 else data[-9:]
            raise PyboardError('timeout waiting for first EOF reception',data)
        data = data[:-1]

        # wait for error output
        data_err = self.read_until(1, b'\x04', timeout=timeout)
        if not data_err.endswith(b'\x04'):
            data_err = data_err if len(data_err)<10 else data_err[-9:]
            raise PyboardError('timeout waiting for second EOF reception', data_err)
        data_err = data_err[:-1]

        # return normal and error output
        return data, data_err

    def exec_raw_no_follow(self, command):
        if isinstance(command, bytes):
            command_bytes = command
        else:
            command_bytes = bytes(command, encoding='utf8')

        # check we have a prompt
        data = self.read_until(1, b'>')
        if not data.endswith(b'>'):
            data = data if len(data)<10 else data[-9:]
            raise PyboardError('could not enter raw repl(no >)',data)

        # write command
        for i in range(0, len(command_bytes), 256):
            self.serial.write(command_bytes[i:min(i + 256, len(command_bytes))])
            time.sleep(0.01)
        self.serial.write(b'\x04')
        # import logging
        # logging.basicConfig(filename='log.txt', level=logging.DEBUG,  format=' %(asctime)s - %(levelname)s- %(message)s')
        # check if we could exec command
        data = self.serial.read(2)
        # logging.debug(data)

        if data != b'OK':
            raise PyboardError('could not exec command',command)

    def exec_raw(self, command, timeout=500, data_consumer=None):
        self.exec_raw_no_follow(command)
        return self.follow(timeout, data_consumer)

    def eval(self, expression):
        ret = self.exec_('print({})'.format(expression))
        ret = ret.strip()
        return ret

    def exec_(self, command, stream_output=False, timeout=500):
        data_consumer = None
        if DEBUG: print("\033[1;35m exec_:\r\n{}\033[0m".format(command))
        if stream_output:
            data_consumer = stdout_write_bytes
        ret, ret_err = self.exec_raw(command, timeout=timeout, data_consumer=data_consumer)
        if ret_err:
            raise PyboardError('exception', ret, ret_err)
        return ret

    def execfile(self, filename, stream_output=False):
        with open(filename, 'rb') as f:
            pyfile = f.read()
        return self.exec_(pyfile, stream_output=stream_output)

    def get_time(self):
        t = str(self.eval('pyb.RTC().datetime()'), encoding='utf8')[1:-1].split(', ')
        return int(t[4]) * 3600 + int(t[5]) * 60 + int(t[6])

# in Python2 exec is a keyword so one must use "exec_"
# but for Python3 we want to provide the nicer version "exec"
setattr(Pyboard, "exec", Pyboard.exec_)

def execfile(filename, device='/dev/ttyACM0', baudrate=115200, user='micro', password='python'):
    pyb = Pyboard(device, baudrate, user, password)
    pyb.enter_raw_repl()
    output = pyb.execfile(filename)
    stdout_write_bytes(output)
    pyb.exit_raw_repl()
    pyb.close()

def main():
    import argparse
    cmd_parser = argparse.ArgumentParser(description='Run scripts on the pyboard.')
    cmd_parser.add_argument('--device', default='/dev/ttyACM0', help='the serial device or the IP address of the pyboard')
    cmd_parser.add_argument('-b', '--baudrate', default=115200, help='the baud rate of the serial device')
    cmd_parser.add_argument('-u', '--user', default='micro', help='the telnet login username')
    cmd_parser.add_argument('-p', '--password', default='python', help='the telnet login password')
    cmd_parser.add_argument('-c', '--command', help='program passed in as string')
    cmd_parser.add_argument('-w', '--wait', default=0, type=int, help='seconds to wait for USB connected board to become available')
    cmd_parser.add_argument('--follow', action='store_true', help='follow the output after running the scripts [default if no scripts given]')
    cmd_parser.add_argument('files', nargs='*', help='input files')
    args = cmd_parser.parse_args()

    def execbuffer(buf):
        try:
            pyb = Pyboard(args.device, args.baudrate, args.user, args.password, args.wait)
            pyb.enter_raw_repl()
            ret, ret_err = pyb.exec_raw(buf, timeout=None, data_consumer=stdout_write_bytes)
            pyb.exit_raw_repl()
            pyb.close()
        except PyboardError as er:
            print(er)
            sys.exit(1)
        except KeyboardInterrupt:
            sys.exit(1)
        if ret_err:
            stdout_write_bytes(ret_err)
            sys.exit(1)

    if args.command is not None:
        execbuffer(args.command.encode('utf-8'))

    for filename in args.files:
        with open(filename, 'rb') as f:
            pyfile = f.read()
            execbuffer(pyfile)

    if args.follow or (args.command is None and len(args.files) == 0):
        try:
            pyb = Pyboard(args.device, args.baudrate, args.user, args.password, args.wait)
            ret, ret_err = pyb.follow(timeout=None, data_consumer=stdout_write_bytes)
            pyb.close()
        except PyboardError as er:
            print(er)
            sys.exit(1)
        except KeyboardInterrupt:
            sys.exit(1)
        if ret_err:
            stdout_write_bytes(ret_err)
            sys.exit(1)

if __name__ == "__main__":
    main()


import sys

def main():
    from threading import Thread
    from .klink import run
    from thonny.web_conn import Webtty

    from flask import Flask,send_from_directory,abort,make_response

    from gevent import pywsgi
    from geventwebsocket.handler import WebSocketHandler

    sysout = sys.stdout
    app = Flask(__name__)
    sockets = Webtty()
    

    def after_request(resp):
        resp.headers['Access-Control-Allow-Origin'] = '*'
        resp.headers['Access-Control-Allow-Private-Network'] = "true"
        return resp

    app.after_request(after_request)
    run(app, sockets, '/hardware')

    appPing = Flask(__name__)
    def ping():
        return '''{"data": {"auto": true}}'''
    appPing.add_url_rule('/ping', 'ping', ping)
    appPing.after_request(after_request)

    wsTh = Thread(target=sockets.listen, )
    wsTh.start()
    
    pingServer = pywsgi.WSGIServer(('0.0.0.0', 55820), appPing)
    print("web server start ... ")
    pingServer.serve_forever()



if __name__ == "__main__":
    main()
    
    





import platform
from .ExecThread import *
from .Uf2Manager import *
from .SerialCom import serialList, serialCom
from .uflash import getDisk
from .ImageManager import saveToBmp
from .kerror import KERR
import shutil,os,time


class HardwareHandler:
  def __init__(self, ws, extensions, userPath, dbgTag=""):
    self.closed = False
    self.ws = ws
    self.extensions = extensions
    self.ext = self.extensions['meow32']
    self.comm = None
    self.commType = None
    self.commList = []
    self.uploadTh = None
    self.pid = 0
    self.params = []
    self.userPath = userPath
    self.usbStatus = []
    self.meow32_ID = "USB\VID_F055&PID_9800&MI_01"
    self.mewbit_ID = "USB\VID_28E9&PID_018A"
    self.dbgTag = dbgTag

  def sendResp(self, pid, result, error=None):
      """发送response"""
      res = {
          "jsonrpc":"2.0",
          "id": pid,
          "result": result
      }
      if error:
          res['error'] = error
      try:
        if not self.closed:
          logging.info("[{}]<<< {}".format(self.dbgTag, json.dumps(res)))
          self.ws.send(json.dumps(res))
      except Exception as err:
        print(err)

  def sendReq(self,method, params={}):
      """发送request"""
      req = {
          "jsonrpc":"2.0",
          "method":method,
          "params": params
      }
      try:
        if not self.closed:
          logging.info("[{}]<<< {}".format(self.dbgTag, json.dumps(req)))
          self.ws.send(json.dumps(req))
      except Exception as err:
        print(err)

  def commRx(self,msg, dt):
      """发送串口读到的数据"""
      if msg == None and dt == -1:
          self.sendReq('connclose')
      else:
          b64 = str(base64.b64encode(msg), 'utf8')
          self.sendReq("data", {"data": b64})

  def handle(self,message):
      """
      供外部使用的方法
      处理收到消息的method/params/id
      根据收到消息的method执行名称对应的方法
      """
      obj = json.loads(message)
      logging.info("[{}]>>> {}".format(self.dbgTag, message))
      if 'id' in obj:
          self.pid = obj['id']
      if 'params' in obj:
          self.params = obj['params']
      method = obj['method']
      method = method.replace("-", "_")
      func = getattr(self, method, None)
      func()

  def sync(self):
    """与前端连接建立收到第一条消息"""
    extId = self.params['extensionId']
    self.ext = self.extensions[extId]
    self.sendResp(self.pid, self.ext)
  
  def install_driver(self):
    if os.name == 'nt':
      try:
        # 1. download filter driver installer
        import tempfile,requests
        tmp = tempfile.mkdtemp()
        url = "http://kittenbot.gitee.io/xes_deploy/kittenbot_filterdrv_1.0.2.0.exe"
        if 'url' in self.params:
          url = self.params['url']
        r = requests.get(url, verify=False)
        if r.ok:
          totalSize = int(r.headers.get('content-length', 0))
          tmpPath = os.path.join(tmp, os.path.basename(url))
          with open(tmpPath, 'wb') as f:
            dataWritten = 0
            for data in r.iter_content(16384):
              f.write(data)
              dataWritten += len(data)
              self.sendReq("upload-status", {"text": "downloading installer".format(dataWritten, totalSize), "progress": int(dataWritten/totalSize*100)})
          n = os.system(tmpPath + " /S")
          self.sendResp(self.pid, {"status": 'ok', 'code': n})
        else:
          self.sendResp(self.pid, None, {"error": "installer download fail", "code": KERR.DRIVER_INSTALL_DOWNLOADFAIL})
        # 2. exec installer
      except Exception as err:
        self.sendResp(self.pid, None, {"error": str(err), "code": KERR.DRIVER_INSTALL_FAIL})
    else:
      self.sendResp(self.pid, None, error={"msg":"No Driver install need for osx", "code":KERR.DRIVER_INSTALL_OSX})

  def listdevice(self):
    """查看可用串口列表"""
      # 1. 检测盘符
    disk = getDisk("PYBFLASH", "2M")
    diskboot = getDisk("ARCADE-F4")
    if diskboot:
      self.sendResp(self.pid, None, error={"msg":"in bootloader mode", "code":KERR.DISK_IN_BOOTLOADER_MODE})
      return
    if platform.system() == "Windows":
      try:
        import win32com.client
        # 添加设备信息到列表
        # win7不引用下面会报错
        import pythoncom
        pythoncom.CoInitialize ()
        win32 = win32com.client.GetObject("winmgmts:")
        self.usbStatus = []
        for usb in win32.InstancesOf("win32_pnpentity"):
          if usb.PNPDeviceID.find(self.meow32_ID) > -1 or usb.PNPDeviceID.find(self.mewbit_ID) > -1:
              self.usbStatus.append({
                "name":usb.Name,
                "errCode":usb.ConfigManagerErrorCode,
                "usbsta":usb.Status
              })

        print(self.usbStatus)
        if len(self.usbStatus):
          for item in self.usbStatus:
            if item["usbsta"] == 'OK':
              # 串口正常，准备连接
              break
            else:
              self.sendResp(self.pid, None, error = {"msg": self.usbStatus, "code": KERR.CANNOT_FIND_COMMPORT_BUTDISK})
        else:
          # 没有设备
          self.sendResp(self.pid, None, error = {"msg": "no device", "code": KERR.CANNOT_FIND_COMMPORT})  
      except Exception as ex:
        # 调用windows中的接口报错
        serPorts = serialList()
        if not serPorts and disk:
          # 如果没有串口但是有盘符
          self.sendResp(self.pid, None, error={
              "msg":"find disk but no comm port", 
              "code":KERR.CANNOT_FIND_COMMPORT_BUTDISK,
              "platform": platform.system(), 
              "release": platform.release(),
              "windowsAPIErr": str(ex)
            })
          return
        elif serPorts:
          # 正常流程，有通信端口
          self.commList = serPorts
          self.sendResp(self.pid, self.commList)
          # print('>>>',ex)
          return
        else:
          # 没有串口,也没盘符,提示用户插入硬件 (esp32同样有效)
          self.sendResp(self.pid, None, error={"msg":"No comm port", "code":KERR.CANNOT_FIND_COMMPORT,"windowsAPIErr": str(ex)})
          return



      serPorts = serialList()
      self.commList = serPorts
      self.sendResp(self.pid, self.commList)
    else:
      serPorts = serialList()
      self.commList = serPorts
      self.sendResp(self.pid, self.commList)

  def connect(self):
    """连接串口"""
    peripheralId = self.params['peripheralId']
    port = None
    if '@' in peripheralId:
      peripheralId = peripheralId.split('@')[0]
    for p in self.commList:
        if p['peripheralId'] == peripheralId:
            port = p
            break
    try:
      if not port:
        raise Exception("无法找到串口 {}".format(peripheralId))
      if port['pid'] == 38912: # meowbit stm32
        if self.ext['id'] != 'meowbit':
          self.ext = self.extensions['meowbit']
      else:
        if self.ext['id'] != 'meow32':
          self.ext = self.extensions['meow32']
      
      if port['type'] == 'serial':
          self.comm = serialCom(self.commRx, self.dbgTag)
          self.comm.connect(p['peripheralId'], baud=1000000)
          self.commType = 'serial'
      self.sendResp(self.pid, p)
    except Exception as e:
      msg = str(e)
      if 'PermissionError' in msg:
        msg = '通信端口被占用，请检查是否有其他页面已经打开了掌机'
      err = {
          "msg": msg,
          "code": KERR.SERIAL_COMM_ERROR
      }
      self.sendResp(self.pid, None, error=err)
    finally:
      print("connected", self.comm)

  def disconnect(self):
    if self.comm :
      self.comm.close()
      
  def write(self):
    """通过串口写数据"""
    msg = self.params['data']
    msg = base64.b64decode(msg)
    # print(msg)
    if self.comm:
        self.comm.write(msg)


import logging
from threading import Lock
from .HardwareHandlerFactory import HardwareHandlerFactory
from .ExecThread import *
from .Uf2Manager import *
from .SerialCom import serialList, serialCom
from .uflash import getDisk
from .ImageManager import saveToBmp
from .kerror import KERR

handler = None
mutex = Lock()

try:
    logging.basicConfig(
        format='%(asctime)s.%(msecs)03d %(message)s',
        level=logging.INFO,
        datefmt='%H:%M:%S')
    import coloredlogs
    coloredlogs.install()
except:
    pass

def handle(websocket, extensions, userPath):
    global handler
    ws = websocket
    import random, string
    letters = string.ascii_lowercase
    instanceTag = ''.join(random.choice(letters) for i in range(6))
    if handler:
        logging.error("new sock without close %s" %handler.dbgTag)
        handler.closed = True
    mutex.acquire()
    logging.warn("\033[1;32m New socket %s %s" %(ws, instanceTag))

    hardwareHandlerFactory = HardwareHandlerFactory()
    handler = hardwareHandlerFactory.handle('meowbit', websocket, extensions, userPath, instanceTag)
    handler.ws = ws
    while not ws.closed and not handler.closed:
        # blocking receive
        if not ws._link_messages.empty():
            message = ws._link_messages.get()
        # if not message:
        #     continue
            try:
                handler.handle(message)
            except Exception as err:
                import traceback
                traceback.print_exc()
                handler.sendReq('ws-error', {'code': KERR.WEBSOCKET_ERR, 'msg': str(err)})
                continue
        else:
            time.sleep(0.05)
    logging.warn("WS connection closed %s" %instanceTag)
    if handler:
        handler.disconnect()
    mutex.release()
    handler = None
    
  


from .ExecThread import *
from .Uf2Manager import *
from .SerialCom import serialList, serialCom
from .uflash import getDisk
from .ImageManager import saveToBmp
from .HardwareHandler import HardwareHandler
from .EspToolMng import uploadEsp32Firm
from .pyboard import Pyboard, PyboardError
from .ampy import Files
from .kerror import KERR
import shutil
import tempfile
from datetime import datetime, timedelta
from functools import wraps
from hashlib import md5
from io import BytesIO
import ast,textwrap
import re
from urllib.parse import urlparse

t0=datetime.now()
def throttle(ms):
    throttle_period = timedelta(milliseconds=ms)
    def throttle_decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            global t0
            now = datetime.now()
            if now - t0 > throttle_period:
                t0 = now
                return fn(*args, **kwargs)
        return wrapper
    return throttle_decorator

IMGRX = "[^\s]+(\.(?i)(jpg|png|bmp))$"

class MeowbitHandler(HardwareHandler):
  def __init__(self, websocket, extensions, userPath, dbgTag=""):
    HardwareHandler.__init__(self, websocket, extensions, userPath, dbgTag)
    
  def list_file_pyb(self):
    self.comm.setPybMutex(True)
    # todo: add pyusb support for ampy
    try:
      pyb = Pyboard(self.comm)
      fp = Files(pyb)
      allFiles = fp.fsmngls()
      self.sendResp(self.pid, {'files': allFiles})
    except (Exception,PyboardError) as e:
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.LIST_FILE_FAIL})
    finally:
      self.comm.setPybMutex(False)

  def list_file(self):
    if self.ext['pyb'] or 'pyb' in self.params:
      return self.list_file_pyb()
    os.path.getsize
    dest = getDisk(self.ext['thumbdisk'],"2M")
    if dest:
        fileFilter = None
        if 'filter' in self.params:
            fileFilter = self.params['filter']
        files = os.listdir(dest)
        allFiles = []
        for f in files:
            fileExtension = os.path.splitext(f)[1]
            if fileFilter:
                if fileExtension in fileFilter:
                    allFiles.append(f)
            else:
                allFiles.append(f)
        self.sendResp(self.pid, {'files': allFiles})
    else:
        self.sendResp(self.pid, None, {"error": "cannot find disk", "code": KERR.LIST_FILE_FAIL})

  def disk_info_pyb(self):
    self.comm.setPybMutex(True)
    # todo: add pyusb support for ampy
    try:
      pyb = Pyboard(self.comm)
      fp = Files(pyb)
      info = fp.fsinfo()
      self.sendResp(self.pid, info)
    except (Exception,PyboardError) as e:
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.GET_DISK_INFO_FAIL})
    finally:
      self.comm.setPybMutex(False)

  def disk_info(self):
    if self.ext['pyb'] or 'pyb' in self.params:
      return self.disk_info_pyb()
    dest = getDisk(self.ext['thumbdisk'],"2M")
    if dest:
        total, used, free = shutil.disk_usage(dest)
        self.sendResp(self.pid, {'total': total, 'used': used, 'free': free})

  def upload_file_pyb(self, params):
    try:
      content = base64.b64decode(params['content'])
      fileName = params['fileName']
      self.comm.setPybMutex(True)
      # todo: add pyusb support for ampy
      pyb = Pyboard(self.comm)
      fp = Files(pyb)
      fp.putasync(fileName, content, delete=1)
      self.sendResp(self.pid, {'status': 'ok', 'file': fileName})
    except (Exception,PyboardError) as e:
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.UPLOAD_IMAGE_FAIL, 'file': fileName})
    finally:
      self.comm.setPybMutex(False)

  def upload_file(self):
    if self.ext['pyb'] or 'pyb' in self.params:
      return self.upload_file_pyb(self.params)

    dest = getDisk(self.ext['thumbdisk'],"2M")
    if not dest:
        self.sendResp(self.pid, None, {"error": 'no_device', "code": KERR.CANNOT_FIND_DISK})
        return
    fileName = self.params['fileName']
    destPath = os.path.join(dest, fileName)
    total, used, free = shutil.disk_usage(dest)
    if free < 100*1024:
        self.sendResp(self.pid, None, {"error": "not enough space", "code": 103, 'file': destPath})
        return
    try:
        with open(destPath, 'wb') as output:
            # logger.info("copy %s to %s" %(fileName, destPath))
            content = base64.b64decode(self.params['content'])
            output.write(content)
            output.flush()
            os.fsync(output) # fix for windows flush with a quantity delay
            self.sendResp(self.pid, {'status': 'ok', 'file': destPath})
    except Exception as e:
        # self.sendResp(self.pid, None, {"error": str(e), "code": 103, 'file': destPath})
        self.sendResp(self.pid, None, {"error": str(e), "code": KERR.UPLOAD_IMAGE_FAIL, 'file': destPath})

  def upload_image_pyb(self, params):
    base64Input = True if 'base64' in params and params['base64'] else False
    fmt = params['fmt'] if 'fmt' in params else 'png'
    fileName = params['fileName']
    if fileName.find('.') > 0:
      fileName = fileName[0:fileName.find('.')]
    width = None
    if 'width' in params:
      width = int(params['width'])
    try:
      imageIo = BytesIO()
      saveToBmp(params['content'], imageIo, width, fmt=fmt, base64Input=base64Input)

      self.comm.setPybMutex(True)
      # todo: add pyusb support for ampy
      pyb = Pyboard(self.comm)
      fp = Files(pyb)
      savePath = fileName+'.'+fmt

      @throttle(100)
      def cb(num, total):
          self.sendReq("upload-status", {"text": "[%s] %s/%s" %(savePath, num, total)})
      fp.put(savePath, imageIo.getvalue(), cb)
      self.sendResp(self.pid, {'status': 'ok', 'file': fileName})
    except (Exception,PyboardError) as e:
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.UPLOAD_IMAGE_FAIL, 'file': savePath})
    finally:
      self.sendReq("upload-status", {})
      self.comm.setPybMutex(False)


  def upload_image(self):
    if self.ext['pyb'] or 'pyb' in self.params:
      return self.upload_image_pyb(self.params)
    dest = getDisk(self.ext['thumbdisk'],"2M")
    fileName = self.params['fileName']
    if not dest:
      self.sendResp(self.pid, None, {"error": 'no_device', "code": KERR.CANNOT_FIND_DISK})
      return
    destPath = os.path.join(dest, fileName)
    total, used, free = shutil.disk_usage(dest)
    if free < 100*1024:
      self.sendResp(self.pid, None, {"error": "not enough space", "code": 103, 'file': destPath})
      return
    try:
      width = None
      if 'width' in self.params:
          width = int(self.params['width'])
      destPath = os.path.splitext(destPath)[0]+'.bmp'
      saveToBmp(self.params['content'], destPath, width)
      self.sendResp(self.pid, {'status': 'ok', 'file': destPath})
    except Exception as e:
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.UPLOAD_IMAGE_FAIL, 'file': destPath})
  

  def delete_file_pyb(self, params):
    try:
      fileName = params['fileName']
      # todo: add pyusb support for ampy
      self.comm.setPybMutex(True)
      pyb = Pyboard(self.comm)
      fp = Files(pyb)
      fp.rm(fileName)
      self.sendResp(self.pid, {'status': 'ok'})
    except (Exception,PyboardError) as e:
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.UPLOAD_IMAGE_FAIL, 'file': fileName})
    finally:
      self.comm.setPybMutex(False)

  def delete_file(self):
    if self.ext['pyb'] or 'pyb' in self.params:
      return self.delete_file_pyb(self.params)
    dest = getDisk(self.ext['thumbdisk'],"2M")
    ret = -1
    if dest:
      fileName = self.params['fileName']
      filePath = os.path.join(dest, fileName)
      if os.path.exists(filePath):
        print(filePath)
        os.remove(filePath)
        ret = 0
    self.sendReq("extension-method", {"extensionId": self.ext['id'], "func": "onDelDone", "msg": ret})   

  def contrastFile(self,boardfile,filedict):
    boardfile_hash = {}
    filedict_hash = {}
    for file_name in boardfile:
      boardfile_hash[boardfile[file_name]["hash"]] = file_name
    for file_name in filedict:
      filedict_hash[filedict[file_name]["hash"]] = file_name

    fileToCopy = list(set(filedict_hash).difference(boardfile_hash))
    fileToDelete = list(set(boardfile_hash).difference(filedict_hash))
    sameFile = list(set(filedict_hash).intersection(boardfile_hash))

    return fileToCopy,sameFile,fileToDelete,boardfile_hash,filedict_hash

  def sync_files(self):
    if self.ext['pyb'] or 'pyb' in self.params:
      return self.sync_files_pyb(self.params)
    try:
      skipList = ['main.py',
                  'VERSION',
                  'filedict.json',
                  'cfg.json',
                  'kodo.gif',
                  'boot.py',
                  'pybcdc.inf',
                  'README.txt',
                  'System Volume Information',
                  '.fseventsd']
      dest = getDisk(self.ext['thumbdisk'],"2M")
      # 传入的json
      filedict = self.params['filedict']
      files = os.listdir(dest)
      # needUpdate = False
      # 掌机内的json
      boardfile = {}
      import platform
      if platform.system() == "Windows":
        try:
          cmdStat = os.system('Chkdsk /f {0}'.format(dest))  # 防止因文件损坏导致的问题
        
        # print('fix hard',cmdStat)
          if cmdStat == 1:  # 需要修复返回代码1，正常返回代码0
            time.sleep(0.5)
        except:
          pass

      if os.path.isfile(os.path.join(dest, 'filedict.json')):
        try:
          fp = open(os.path.join(dest, 'filedict.json'))
          boardfile = json.load(fp)
        except Exception as erro:
          import traceback
          traceback.print_exc()
          for bdfile in files:
            if bdfile not in skipList:
              filePath = os.path.join(dest, bdfile)
              try:
                os.remove(filePath)
              except:
                import traceback
                traceback.print_exc()
        finally:
          fp.close()
          os.remove(os.path.join(dest, 'filedict.json'))
      else:
        for bdfile in files:
          if bdfile not in skipList:
            filePath = os.path.join(dest, bdfile)
            try:
              os.remove(filePath)
            except:
              import traceback
              traceback.print_exc()
      realRootFile = os.listdir(os.path.join(dest))
      fileToCopy,sameFile,fileToDelete,boardfile_hash,filedict_hash = self.contrastFile(boardfile,filedict)

      # 删除掌机上不在资源包中的文件
      for deleteHash in fileToDelete:
        fileName = boardfile_hash[deleteHash]
        if (fileName in skipList) or fileName.startswith('.'):
          continue
        if re.search(IMGRX, fileName):
          fileName = os.path.splitext(fileName)[0]+'.bmp'
        if fileName in realRootFile:
          fpath = os.path.join(dest, fileName)
          os.remove(fpath)
          self.sendReq("upload-status", {"text": "delete file: %s" %(fileName)})
          # needUpdate = True
      # copy files to target
      for copyHash in fileToCopy:
        
        fname = filedict_hash[copyHash]
        f = filedict[fname]
        if 'content' in f:
          content = f['content']
          del filedict[fname]['content']
        elif 'base64' in f:
          content = base64.b64decode(f['base64'])
          del filedict[fname]['base64']
        elif 'url' in f:
          r = requests.get(f['url'], verify=False)
          del filedict[fname]['url']
          if r.ok:
            content = r.content
          else:
            raise RuntimeError("download %s failed" %f['url'])
          # todo: support url download format
        if re.search(IMGRX, fname):
          fielIo = BytesIO(content)
          imageIo = BytesIO()
          saveToBmp(fielIo, imageIo, fmt='bmp')
          content = imageIo.getvalue()
          filedict[fname]['saved'] = len(content)
          fname = os.path.splitext(fname)[0]+'.bmp'
        # todo: support image write
        # todo: stm32 use direct file copy
        with open(os.path.join(dest, fname), 'wb') as fp:
          fp.write(content)
          self.sendReq("upload-status", {"text": "%s >> %s" %(fname, dest)})
          os.fsync(fp)
        # needUpdate = True

      for sameHash in sameFile:
        fname = filedict_hash[sameHash]
        try:
          del filedict[fname]['base64']
        except:
          del filedict[fname]['url']
      # if needUpdate:
      with open(os.path.join(dest, 'filedict.json'), 'w') as fp:
        json.dump(filedict, fp, indent=4)
        os.fsync(fp)
      self.sendResp(self.pid, {'status': 'ok'})
    except (Exception,PyboardError) as e:
      import traceback
      tb_err = traceback.format_exc()
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.SYNC_FILE_FAIL, "traceback": tb_err})
    finally:
      self.sendReq("upload-status", {})
      self.comm.setPybMutex(False)

  def _getContent(self, fname, f):
    if 'content' in f:
      content = f['content']
    elif 'base64' in f:
      content = base64.b64decode(f['base64'])
    elif 'url' in f:
      # cdn hash 有延时
      retryCnt = 3
      url = f['url']
      while retryCnt:
        r = requests.get(url, verify=False)
        if r.ok:
          content = r.content
          break
        else:
          retryCnt -= 1
          time.sleep(1)
          if retryCnt == 0:
            raise RuntimeError("download %s failed" %url)
    if re.search(IMGRX, fname):
      fielIo = BytesIO(content)
      imageIo = BytesIO()
      saveToBmp(fielIo, imageIo, fmt='png')
      content = imageIo.getvalue()
      fname = os.path.splitext(fname)[0]+'.png'
      calcMd5 = md5(imageIo.getbuffer()).hexdigest()
    else:
      calcMd5 = md5(content).hexdigest()
    
    return (fname, content, calcMd5)

  def sync_files_pyb(self, params):
    try:
      d = self.params['filedict']
      rst = False
      if 'reset' in self.params:
        rst = self.params['reset']
      self.comm.setPybMutex(True)
      pyb = Pyboard(self.comm)
      fp = Files(pyb)

      _content = {}
      _newFileDict = {}
      
      for _key in d:
        f = d[_key]
        (fname, content, calcMd5) = self._getContent(_key, f)
        # update hash from 
        _newFileDict[fname] = {
          'hash': '{}_{}'.format(fname, calcMd5),
          'size': len(content)
        }
        _content[fname] = content
        
      # compare file 
      _uploadFileDict = fp.syncCheck(_newFileDict, reset=rst)
      # do upload
      @throttle(100)
      def cb(num, total):
        self.sendReq("upload-status", {"text": "[%s] %s/%s" %(fname, num, total)})
      for fname in _uploadFileDict:
        fp.putasync(fname, _content[fname], cb)
      fp.showmsg("文件同步完成啦",title='文件同步',type='succ')
      self.sendResp(self.pid, {'status': 'ok'})
    except (Exception,RuntimeError,PyboardError) as e:
      import traceback
      tb_err = traceback.format_exc()
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.SYNC_FILE_FAIL, "traceback": tb_err})
    finally:
      self.sendReq("upload-status", {})
      self.comm.setPybMutex(False)

  def normal_repl(self):
    try:
      self.comm.setPybMutex(True)
      pyb = Pyboard(self.comm)
      pyb.enter_normal_repl()
      self.sendResp(self.pid, {'status': 'ok'})
    except (Exception,PyboardError) as e:
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.GO_NORMAL_REPL_FAIL})
    finally:
      self.comm.setPybMutex(False)

  def uname_info(self):
    try:
      self.comm.setPybMutex(True)
      pyb = Pyboard(self.comm)
      fp = Files(pyb)
      info = fp.uname()
      self.sendResp(self.pid, {'info': info})
    except (Exception,PyboardError) as e:
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.GO_NORMAL_REPL_FAIL})
    finally:
      self.comm.setPybMutex(False)

  def run_code(self):
    try:
      code = self.params['code']
      self.comm.setPybMutex(True)
      pyb = Pyboard(self.comm)
      ampy = Files(pyb)
      ret = ampy.run_code(code)
      self.sendResp(self.pid, {'ret': ret})
    except (Exception,PyboardError) as e:
      self.sendResp(self.pid, None, {"error": str(e), "code": KERR.EXEC_COMMAND_FAIL})
    finally:
      self.comm.setPybMutex(False)

  def upload_firmware(self):
    if self.ext["fwtype"] == "uf2":
        # check if in bootloader
        # dest = getDisk(self.ext['thumbdisk'],"2M")
        dest = getDisk(self.ext['thumbdisk'])

        if dest:
          for i in range(5):
            try:
              self.comm.write(b'import machine\r\nmachine.bootloader()\r\n')
            except:
              pass
            time.sleep(2)
            dest_boot = getDisk('ARCADE-F4')
            if dest_boot:
              break      #  进入bootload成功，执行后面的步骤
          if not dest_boot:
            self.sendResp(self.pid, None, {"error": 'enter bootloader faid', "code": KERR.UPLOAD_FIRM_FAIL})
        def onDone(msg, err):
            if err:
                self.sendResp(self.pid, None, {"error": msg, "code": KERR.UPLOAD_FIRM_FAIL})
            else:
                self.sendResp(self.pid, {'status': 'ok', 'msg': msg})
        uploadUf2(self.ext, self.params, self.sendReq, self.userPath, onDone)
    elif self.ext["fwtype"] == "esptool" or self.ext["fwtype"] == "esp":
        port=None
        # todo: add pyusb firmware retore
        if self.comm and self.comm.opened:
            tmp = tempfile.mkdtemp()
            if 'url' in self.params:
                url = self.params['url']
                _parsedUrl = urlparse(url)
                self.sendReq("upload-status", {"text": "Downloading %s" %url})
                r = requests.get(url, verify=False)
                if r.ok:
                  firmPath = os.path.join(tmp, os.path.basename(_parsedUrl.path))
                  with open(firmPath, 'wb') as f:
                      f.write(r.content)
                else:
                  self.sendResp(self.pid, None, {"error": "Firmware download fail", "code": KERR.UPLOAD_FIRM_FAIL})
                  return
            self.sendReq('connclose')
            def onLog(msg):
                # print('>>>', msg)
                msg = re.findall(r'[(](.*?)[%]', msg)
                if len(msg) == 0:
                  msg = [0]
                
                self.sendReq("upload-status", {"text": "#", "progress": int(msg[0])})
            def onDone(err):
                self.comm.setPybMutex(False)
                self.sendReq("upload-status", {})
                if err:
                    import traceback
                    traceback.print_exc()
                    self.sendResp(self.pid, None, {"error": str(err), "code": KERR.UPLOAD_FIRM_FAIL})
                else:
                    shutil.rmtree(tmp)
                    self.sendResp(self.pid, {'status':'ok'})
            config = {
                'firmwarePath': firmPath,
                'flashAddr': self.ext['flashAddr'] if 'flashAddr' in self.ext else 0x1000
            }
            if self.comm.ser:
              port=self.comm.ser.name
              self.comm.ser.close()
              uploadEsp32Firm(port, config, onLog, onDone)
            elif self.comm.dev: # filter driver mode
              self.comm.setPybMutex(True)
              uploadEsp32Firm("USB_CDC", config, onLog, onDone, self.comm)
            else:
              raise Exception("No port selected for esptool")
        else:
          self.sendResp(self.pid, None, {"error": "no download port", "code": KERR.UPLOAD_FIRM_FAIL})


from setuptools import setup,find_packages
setup(name='pylink_core',
      version='0.5.11',
      description='a small example',
      classifiers=[
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
    ],
      url='https://www.python.org/',
      install_requires=[
        'pyusb==1.0.2'
      ],
      author='zhushaodong123',
      author_email='',
      license='NEU',
      packages=find_packages(),
      package_data={"":["*.pyc"]},
      zip_safe=True
     )




class KERR:
    WEBSOCKET_ERR=100
    SERIAL_COMM_ERROR=101
    CANNOT_FIND_DISK=102
    UPLOAD_IMAGE_FAIL=103


    UPLOAD_FIRM_FAIL=110
    CANNOT_FIND_COMMPORT=170
    CANNOT_FIND_COMMPORT_BUTDISK=171
    DISK_IN_BOOTLOADER_MODE=172
    GO_NORMAL_REPL_FAIL=175
    EXEC_COMMAND_FAIL=176
    DRIVER_INSTALL_FAIL = 180
    DRIVER_INSTALL_DOWNLOADFAIL = 181
    DRIVER_INSTALL_OSX = 183
    LIST_FILE_FAIL=150
    GET_DISK_INFO_FAIL=160

    SYNC_FILE_FAIL=165


    CANNT_USE_GETOBJECT = 189
    UNKNOW_ERROR = 188





# Adafruit MicroPython Tool - File Operations
# Author: Tony DiCola
# Copyright (c) 2016 Adafruit Industries
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import ast
import json
import textwrap
import binascii
from .pyboard import PyboardError


BUFFER_SIZE = 1024  # Amount of data to read or write to the serial port at a time.
# This is kept small because small chips and USB to serial
# bridges usually have very small buffers.


class DirectoryExistsError(Exception):
    pass


class Files(object):
    """Class to interact with a MicroPython board files over a serial connection.
    Provides functions for listing, uploading, and downloading files from the
    board's filesystem.
    """

    def __init__(self, pyboard):
        """Initialize the MicroPython board files class using the provided pyboard
        instance.  In most cases you should create a Pyboard instance (from
        pyboard.py) which connects to a board over a serial connection and pass
        it in, but you can pass in other objects for testing, etc.
        """
        self._pyboard = pyboard

    def get(self, filename):
        """Retrieve the contents of the specified file and return its contents
        as a byte string.
        """
        # Open the file and read it a few bytes at a time and print out the
        # raw bytes.  Be careful not to overload the UART buffer so only write
        # a few bytes at a time, and don't use print since it adds newlines and
        # expects string data.
        command = """
            import sys
            import ubinascii
            with open('{0}', 'rb') as infile:
                while True:
                    result = infile.read({1})
                    if result == b'':
                        break
                    len = sys.stdout.write(ubinascii.hexlify(result))
        """.format(
            filename, BUFFER_SIZE
        )
        self._pyboard.enter_raw_repl()
        try:
            out = self._pyboard.exec_(textwrap.dedent(command))
        except PyboardError as ex:
            # Check if this is an OSError #2, i.e. file doesn't exist and
            # rethrow it as something more descriptive.
            try:
                
                #if ex.args[2].decode("utf-8").find("OSError: [Errno 2] ENOENT") != -1:
                if len(ex.args) > 2 and ex.args[2].decode("utf-8").find("OSError: [Errno 2] ENOENT") != -1:
                    raise RuntimeError("No such file: {0}".format(filename))
                else:
                    raise ex
            except UnicodeDecodeError:
                raise ex
        self._pyboard.exit_raw_repl()
        return binascii.unhexlify(out)

    def ls(self, directory=None, long_format=True, bundle_format=False, recursive=False):
        """List the contents of the specified directory (or root if none is
        specified).  Returns a list of strings with the names of files in the
        specified directory.  If long_format is True then a list of 2-tuples
        with the name and size (in bytes) of the item is returned.  Note that
        it appears the size of directories is not supported by MicroPython and
        will always return 0 (i.e. no recursive size computation).
        """

        # Disabling for now, see https://github.com/adafruit/ampy/issues/55.
        # # Make sure directory ends in a slash.
        # if not directory.endswith("/"):
        #     directory += "/"

        # Make sure directory starts with slash, for consistency.
        if not directory:
            directory = "os.getcwd()"
        else:
            if not directory.startswith("/"):
                directory = "/" + directory
            directory = "'{}'".format(directory)

        command = """\
                try:        
                    import os
                except ImportError:
                    import uos as os\n"""

        if recursive:
            command += """\
                def listdir(directory):
                    result = set()
                    def _listdir(dir_or_file):
                        try:
                            # if its a directory, then it should provide some children.
                            children = os.listdir(dir_or_file)
                        except OSError:                        
                            # probably a file. run stat() to confirm.
                            os.stat(dir_or_file)
                            result.add(dir_or_file) 
                        else:
                            # probably a directory, add to result if empty.
                            if children:
                                # queue the children to be dealt with in next iteration.
                                for child in children:
                                    # create the full path.
                                    if dir_or_file == '/':
                                        next = dir_or_file + child
                                    else:
                                        next = dir_or_file + '/' + child
                                    
                                    _listdir(next)
                            else:
                                result.add(dir_or_file)                     
                    _listdir(directory)
                    return sorted(result)\n"""
        else:
            command += """\
                def listdir(directory):
                    if directory == '/':                
                        return sorted([directory + f for f in os.listdir(directory)])
                    else:
                        return sorted([directory + '/' + f for f in os.listdir(directory)])\n"""

        # Execute os.listdir() command on the board.
        if bundle_format:
            command += """
                r = []
                for f in listdir({0}):
                    size = os.stat(f)[6]                    
                    r.append((f, size))
                print(r)
            """.format(
                directory
            )
        elif long_format:
            command += """
                r = []
                for f in listdir({0}):
                    size = os.stat(f)[6]                    
                    r.append('{{0}} - {{1}} bytes'.format(f, size))
                print(r)
            """.format(
                directory
            ) 
        else:
            command += """
                print(listdir({0}))
            """.format(
                directory
            )
        self._pyboard.enter_raw_repl()
        try:
            out = self._pyboard.exec_(textwrap.dedent(command))
        except PyboardError as ex:
            import traceback
            traceback.print_exc()
            print("PyboardError", ex)
            # Check if this is an OSError #2, i.e. directory doesn't exist and
            # rethrow it as something more descriptive.
            if ex.args[2].decode("utf-8").find("OSError: [Errno 2] ENOENT") != -1:
                raise RuntimeError("No such directory: {0}".format(directory))
            else:
                raise ex
        self._pyboard.exit_raw_repl()
        # Parse the result list and return it.
        return ast.literal_eval(out.decode("utf-8"))

    def fsmngls(self):
        # only list user file
        command = """
            import fsmng
            print(fsmng.list())
        """
        self._pyboard.enter_raw_repl()
        try:
            out = self._pyboard.exec_(textwrap.dedent(command))
        except PyboardError as ex:
            raise ex
        self._pyboard.exit_raw_repl()
        info = ast.literal_eval(out.decode("utf-8"))
        return info

    def mkdir(self, directory, exists_okay=False):
        """Create the specified directory.  Note this cannot create a recursive
        hierarchy of directories, instead each one should be created separately.
        """
        # Execute os.mkdir command on the board.
        command = """
            try:
                import os
            except ImportError:
                import uos as os
            os.mkdir('{0}')
        """.format(
            directory
        )
        self._pyboard.enter_raw_repl()
        try:
            out = self._pyboard.exec_(textwrap.dedent(command))
        except PyboardError as ex:
            # Check if this is an OSError #17, i.e. directory already exists.
            if ex.args[2].decode("utf-8").find("OSError: [Errno 17] EEXIST") != -1:
                if not exists_okay:
                    raise DirectoryExistsError(
                        "Directory already exists: {0}".format(directory)
                    )
            else:
                raise ex
        self._pyboard.exit_raw_repl()

    def put(self, filename, data, callback=None):
        """Create or update the specified file with the provided data.
        """
        # Open the file for writing on the board and write chunks of data.
        self._pyboard.enter_raw_repl()
        self._pyboard.exec_("f = open('{0}', 'wb')".format(filename))
        size = len(data)
        # Loop through and write a buffer size chunk of data at a time.
        for i in range(0, size, BUFFER_SIZE):
            chunk_size = min(BUFFER_SIZE, size - i)
            chunk = repr(data[i : i + chunk_size])
            # Make sure to send explicit byte strings (handles python 2 compatibility).
            if not chunk.startswith("b"):
                chunk = "b" + chunk
            self._pyboard.exec_("f.write({0})".format(chunk))
            if callback:
                callback(i, size)
        self._pyboard.exec_("f.close()")
        self._pyboard.exit_raw_repl()

    def putasync(self, filename, data, callback=None, delete=False):
        # Open the file for writing on the board and write chunks of data.
        self._pyboard.enter_raw_repl()
        size = len(data)
        if delete:
            cmd = '''
            try:
                os.remove("{}")
            except:
                pass
            '''.format(filename)
            self._pyboard.exec_(textwrap.dedent(cmd))
        # Loop through and write a buffer size chunk of data at a time.
        for i in range(0, size, BUFFER_SIZE):
            chunk_size = min(BUFFER_SIZE, size - i)
            chunk = repr(data[i : i + chunk_size])
            # Make sure to send explicit byte strings (handles python 2 compatibility).
            if not chunk.startswith("b"):
                chunk = "b" + chunk
            if chunk_size + i == size:
                timeout = 5000
            else:
                timeout = 500
            self._pyboard.exec_("import fsmng\nfsmng.putasync('{}',{},{},{})".format(filename,chunk,i,size), timeout=timeout)
            if callback:
                callback(i, size)
        self._pyboard.exit_raw_repl()

    def showmsg(self, msg, title, type='err'):
        # Open the file for writing on the board and write chunks of data.
        self._pyboard.enter_raw_repl()
        self._pyboard.exec_("screen.showmsg('{}','{}', '{}')".format(msg, title, type))
        self._pyboard.exit_raw_repl()

    def uname(self):
        command = """
            try:
                import os
            except ImportError:
                import uos as os
            print(list(os.uname()))
        """
        self._pyboard.enter_raw_repl()
        try:
            out = self._pyboard.exec_(textwrap.dedent(command))
        except PyboardError as ex:
            raise ex
        self._pyboard.exit_raw_repl()
        info = ast.literal_eval(out.decode("utf-8"))
        if len(info) == 6: # meow32
            model = info[0] 
            date = info[4]
            ver = info[4][3] + '.' +info[4][5:7] + info[4][8:10]
            intro = info[5]
            mac = info[1]
            ret = (model, model, ver, ver, date, intro, mac)
            return ret    
        else:
            return info

    def fsinfo(self, prompt='/'):
        """Get filesystem info"""
        command = """
            import fsmng
            print(fsmng.diskinfo())
        """.format(
            prompt
        )
        self._pyboard.enter_raw_repl()
        try:
            out = self._pyboard.exec_(textwrap.dedent(command))
        except PyboardError as ex:
            raise ex
        self._pyboard.exit_raw_repl()
        info = ast.literal_eval(out.decode("utf-8"))
        return {'total': info[0], 'used': info[1], 'free': info[2]}

    def rm(self, filename):
        """Remove the specified file or directory."""
        command = """
            try:
                import os
            except ImportError:
                import uos as os
            os.remove('{0}')
        """.format(
            filename
        )
        self._pyboard.enter_raw_repl()
        try:
            out = self._pyboard.exec_(textwrap.dedent(command))
        except PyboardError as ex:
            message = ex.args[2].decode("utf-8")
            # Check if this is an OSError #2, i.e. file/directory doesn't exist
            # and rethrow it as something more descriptive.
            if message.find("OSError: [Errno 2] ENOENT") != -1:
                raise RuntimeError("No such file/directory: {0}".format(filename))
            # Check for OSError #13, the directory isn't empty.
            if message.find("OSError: [Errno 13] EACCES") != -1:
                raise RuntimeError("Directory is not empty: {0}".format(filename))
            else:
                raise ex
        self._pyboard.exit_raw_repl()

    def resetfs(self):
        """Reset user disk"""
        command = """
            try:
                fp = open("main.py")
                mainpy = fp.read()
                fp.close()
            except:
                mainpy = None
                pass
            import inisetup
            inisetup.resetfs(mainpy)
        """
        self._pyboard.enter_raw_repl()
        try:
            out = self._pyboard.exec_(textwrap.dedent(command))
        except PyboardError as ex:
            raise ex
        self._pyboard.exit_raw_repl()

    def syncCheck(self, d, reset=False):
        if type(d) != str:
            d = json.dumps(d)
        command = """
        import fsmng
        print(fsmng.syncFileCheck('{}',reset={}))
        """.format(d, reset)
        self._pyboard.enter_raw_repl()
        try:
            out = self._pyboard.exec_(textwrap.dedent(command), timeout=10000)
        except PyboardError as ex:
            raise ex
        self._pyboard.exit_raw_repl()
        return ast.literal_eval(out.decode("utf-8"))

    def rmdir(self, directory, missing_okay=False):
        """Forcefully remove the specified directory and all its children."""
        # Build a script to walk an entire directory structure and delete every
        # file and subfolder.  This is tricky because MicroPython has no os.walk
        # or similar function to walk folders, so this code does it manually
        # with recursion and changing directories.  For each directory it lists
        # the files and deletes everything it can, i.e. all the files.  Then
        # it lists the files again and assumes they are directories (since they
        # couldn't be deleted in the first pass) and recursively clears those
        # subdirectories.  Finally when finished clearing all the children the
        # parent directory is deleted.
        command = """
            try:
                import os
            except ImportError:
                import uos as os
            def rmdir(directory):
                os.chdir(directory)
                for f in os.listdir():
                    try:
                        os.remove(f)
                    except OSError:
                        pass
                for f in os.listdir():
                    rmdir(f)
                os.chdir('..')
                os.rmdir(directory)
            rmdir('{0}')
        """.format(
            directory
        )
        self._pyboard.enter_raw_repl()
        try:
            out = self._pyboard.exec_(textwrap.dedent(command))
        except PyboardError as ex:
            message = ex.args[2].decode("utf-8")
            # Check if this is an OSError #2, i.e. directory doesn't exist
            # and rethrow it as something more descriptive.
            if message.find("OSError: [Errno 2] ENOENT") != -1:
                if not missing_okay:
                    raise RuntimeError("No such directory: {0}".format(directory))
            else:
                raise ex
        self._pyboard.exit_raw_repl()

    def run(self, filename, wait_output=True, stream_output=True):
        """Run the provided script and return its output.  If wait_output is True
        (default) then wait for the script to finish and then return its output,
        otherwise just run the script and don't wait for any output.
        If stream_output is True(default) then return None and print outputs to
        stdout without buffering.
        """
        self._pyboard.enter_raw_repl()
        out = None
        if stream_output:
            self._pyboard.execfile(filename, stream_output=True)
        elif wait_output:
            # Run the file and wait for output to return.
            out = self._pyboard.execfile(filename)
        else:
            # Read the file and run it using lower level pyboard functions that
            # won't wait for it to finish or return output.
            with open(filename, "rb") as infile:
                self._pyboard.exec_raw_no_follow(infile.read())
        self._pyboard.exit_raw_repl()
        return out

    def run_code(self, code):
        self._pyboard.enter_raw_repl()
        try:
            code = textwrap.dedent(code)
            out = self._pyboard.exec_(code, timeout=10000)
        except PyboardError as ex:
            raise ex
        self._pyboard.exit_raw_repl()
        return ast.literal_eval(out.decode("utf-8"))

if __name__ == "__main__":
    import pyboard
    import SerialCom as SC
    com = SC.serialList()[0]['peripheralId']
    def testrx(msg, dt):
        print(">>>>>>>>", msg, dt)

    s = SC.serialCom(testrx)
    s.connect(com, baud=115200)

    # setattr(ser, 'pybMutex', False)
    s.setPybMutex(True)
    pyb = pyboard.Pyboard(s.ser)
    fp = Files(pyb)
    ls = fp.ls()
    s.setPybMutex(False)
    print(ls)
